<?
/* --------------------------------------------------------------------
   Copyright (c) 2004-2012 Daniel Pedemont (dan@pedemont.net)
   -------------------------------------------------------------------- */
/**
 * main library
 *
 * generic functions for CLI, utility, curl, http, etc...
 *
 * @link http://dev.pedemont.com/sonic/
 * @link http://dev.pedemont.com/sonic/downloads/src/lib/sonic.inc
 * @package sonic
 * @version 0.2.23
 * @abstract main utility include library
 */

define('SONIC_VERSION', '0.2.30+');

if (defined('SOL_IP'));
else
  define('SOL_IP', 0);

//SCOPE FUNCS, moved to upper scope because of delcare tick
/**
 * PHP extension check (core requirement checker)
 *
 * checks for loaded .dll or .so, will echo on error
 *
 * @param array $exts list of PHP extensions load check
 * @param boolean $fatal if set true will exit on failure
 * @return array array of loaded extensions in list
*/
function sonic_ext_check($exts, $fatal = true)
{
  static	$_sec_warned = false;

  $ext_ok = array();

  if (ini_get('enable_dl'));
  else
  {
    if ($_sec_warned);
    else
    {
      $_sec_warned = true;

      error_log('Warning: sonic_ext_check() cant check if extensions are loaded');
    }

    return $ext_ok;
  }
  
  $ext2 = (substr(PHP_OS, 0, 3) == 'WIN') ? '.dll' : '.so';
  while (list($i, $ext) = each($exts))
  {
    if (extension_loaded($ext))
      continue;

    if (@dl($ext.'.'.$ext2))
      array_push($ext_ok, $ext);
    else
    {
      error_log('FATAL: PHP is missing the required extension '.$ext.$ext2);

      if ($fatal)
        die(1);
    }
  }

  return $ext_ok;
}

/**
 * environment init. (define defaults)
 *
 * sets defined constants and sets missing defaults in supplied sonic_ini() array
 * also sets the default timezone
 *
 * @param array $sonic_ini array of type sonic_ini()
 * @return array array of type sonic_ini()
*/
function sonic_define($sonic_ini = array())
{
  static $set = false;
  if ($set)
    return $sonic_ini;
  $set = true;

  define('SONIC_TMP_PATH', ($sonic_ini['tmp_path'] ? $sonic_ini['tmp_path'] : '/tmp'));

  if ($sonic_ini['date_tz']) {}
  else if ($sonic_ini['date_tz'] = ini_get('date.timezone')) {}
  else
  {
    if (substr(PHP_OS, 0, 3) == 'WIN')
      $sonic_ini['date_tz'] = 'UTC';
    else
    {
      error_log('Notice: SONIC ini \'date_tz\' is not set/ignored, PHP will set its own');

//      $sonic_ini['date_tz'] = 'System/Localtime';

      if (function_exists('date_default_timezone_get'))
        $sonic_ini['date_tz'] = date_default_timezone_get();
      else
        $sonic_ini['date_tz'] = 'UTC';
    }
  }

  if (function_exists('date_default_timezone_set'))
    @date_default_timezone_set($sonic_ini['date_tz']);

  $_ptz = ini_set('date.timezone', $sonic_ini['date_tz']);

  //false/boolean on failure, string on success
  if (gettype($_ptz) == 'string');
  else
  {
    error_log('Warning: SONIC ini \'date_tz\' is required/invalid, setting as \'UTC\'');

    $sonic_ini['date_tz'] = 'UTC';
    ini_set('date.timezone', $sonic_ini['date_tz']);
  }

  define('SONIC_DATE_TZ', $sonic_ini['date_tz']);
  define('SONIC_DATEF', ($sonic_ini['date_format'] ? $sonic_ini['date_format'] : 'M-d H:i:s'));
  define('SONIC_MIME_FILE', ($sonic_ini['mime_file'] ? $sonic_ini['mime_file'] : '/etc/mime.types'));

  //php4 socket settings
  define('SONIC_SOCK_DOMAIN', ($sonic_ini['socket_domain'] ? $sonic_ini['socket_domain'] : AF_INET));
  define('SONIC_SOCK_TYPE', ($sonic_ini['socket_type'] ? $sonic_ini['socket_type'] : SOCK_STREAM));
  define('SONIC_SOCK_PROTOCOL', ($sonic_ini['socket_protocol'] ? $sonic_ini['socket_protocol'] : SOL_TCP));
  //only define scalar, however can use more
//  define('SONIC_SOCK_OPTIONS', ($sonic_ini['socket_options'] ? $sonic_ini['socket_options'] : array(SO_REUSEADDR => 1)));
//  define('SONIC_SOCK_OPTIONS', ($sonic_ini['socket_options'] ? $sonic_ini['socket_options'] : 'REUSEADDR=1')); 
  define('SONIC_SOCK_OPTIONS', ($sonic_ini['socket_options'] ? $sonic_ini['socket_options'] : SO_REUSEADDR)); 

  //TODO beta place holders
  //php5 stream settings (not the same as sonic streams, but php5 stream/socket)
  define('SONIC_STRM_TRANSPORT', ($sonic_ini['strm_transport'] ? strtolower($sonic_ini['strm_transport']) : 'tcp'));
  if ($sonic_ini['strm_svrflags'])
  {
    define('SONIC_STRM_SVRFLAGS', $sonic_ini['strm_svrflags']);
  }
  else if (SONIC_STRM_TRANSPORT == 'udp' OR SONIC_STRM_TRANSPORT == 'udg') //SONIC_SOCK_TYPE = SOCK_DGRAM
  {
    define('SONIC_STRM_SVRFLAGS', STREAM_SERVER_BIND);
  }
  else //if (SONIC_STRM_TRANSPORT == 'tcp' OR SONIC_STRM_TRANSPORT == 'unix') //SONIC_SOCK_TYPE = SOCK_STREAM
  {
    define('SONIC_STRM_SVRFLAGS', STREAM_SERVER_LISTEN);
  }
  //only define scalar, however can use more
  define('SONIC_STRM_OPTIONS', ($sonic_ini['strm_options'] ? $sonic_ini['strm_options'] : '')); 

  $sonic_ini['tmp_path'] 	= SONIC_TMP_PATH;
  $sonic_ini['date_tz'] 	= SONIC_DATE_TZ;
  $sonic_ini['date_format'] 	= SONIC_DATEF;
  $sonic_ini['mime_file']	= SONIC_MIME_FILE;
  $sonic_ini['socket_domain']	= SONIC_SOCK_DOMAIN;
  $sonic_ini['socket_type']	= SONIC_SOCK_TYPE;
  $sonic_ini['socket_protocol']	= SONIC_SOCK_PROTOCOL;
  $sonic_ini['socket_options']	= SONIC_SOCK_OPTIONS;
  $sonic_ini['strm_transport']	= SONIC_SOCK_TRANSPORT;
  $sonic_ini['strm_svrflags']	= SONIC_STRM_SVRFLAGS;
  $sonic_ini['strm_options']	= SONIC_STRM_OPTIONS;
 
  return $sonic_ini;
}

//MAIN
//NB, you can comment out curl if you are
//    sure that you wont use the functions that
//    use it
$exts = array('pcre', 'curl');
sonic_ext_check($exts);

//FUNCS
/**
 * parse_ini_file() SONIC
 *
 * sets global $_SONIC['SONIC_INI'] from SONIC section of ini file
 *
 * @global array will set $_SONIC['SONIC_INI'] array
 * @param string $filename full location to ini file to be parsed
 * @param boolean $process_sections see parse_ini_file()
 * @return array array of type parse_ini_file()
*/ 
function sonic_ini($filename, $process_sections = false)
{
  global $_SONIC;

  if ($loaded_ini = parse_ini_file($filename, $process_sections))
  {
    if ($sonic_ini = $loaded_ini[SONIC])
      $_SONIC['SONIC_INI'] = $sonic_ini;
    else
    {
      $sonic_ini = array();
      $_SONIC['SONIC_INI'] = false;
    }
  }
  else
  {
    $sonic_ini = array();
    $_SONIC['SONIC_INI'] = false;
  }

  return $loaded_ini;
}

/**
 * file_mode_contents() (lock safe)
 *
 * write string to filename
 *
 * @param string $file_name full location to file to be written/appended
 * @param string $string data string to write to file
 * @param string $mode file write mode, eg. w, a, x
 * @param string $lock_dir path to directory used for locking
 * @return boolean true if successful or false at first fail
*/ 
function sonic_file_mode_contents($file_name, $string, $mode = 'w', $lock_dir = SONIC_TMP_PATH)
{
  $sfpc_hash = '';

  $sfpc_lock_file = $lock_dir.'/sonic-sfpc'.$sfpc_hash.'.lock';
//  $sfpc_lock_file = $lock_dir.'/sonic-sfpc'.md5($sfpc_hash).'lock';

  $time_now = time();
  $time_max = 5;
  while (file_exists($sfpc_lock_file))
  {
    $lock_time = filectime($sfpc_lock_file);

    if ($lock_time - $time_now > $time_max)
    {
      unlink($sfpc_lock_file);
      break;
    }

    // wait for 0.25 seconds
    usleep(250000);
  }

  touch($sfpc_lock_file);
  $out = file_mode_contents($file_name, $string, $mode);
  unlink($sfpc_lock_file);

  return $out;
}

if (version_compare(phpversion(), '5.0.0', '<'))
{
/**
 * @see file_puts_contents()
 * @param string $file_name target path to the data file
 * @param string $string the data to write
 * @return boolean see file_put_contents()
*/ 
  function file_put_contents($file_name, $string)
  {
    return file_mode_contents($file_name, $string, 'w');
  }
}

/**
 * file_put_contents() (reserved alias)
 *
 * write string to filename
 *
 * @see file_puts_contents()
 * @param string $file_name full location to file to be written/appended
 * @param string $string data string to write to file
 * @param string $mode file write mode, eg. w, a, x
 * @return boolean true if successful or false at first fail
*/ 
function file_mode_contents($file_name, $string, $mode = 'w')
{
  if ($string);
  else
    return true;

  $mode = strtolower($mode);
  switch ($mode)
  {
    case 'w':
    case 'w+':
    case 'a':
    case 'a+':
    case 'x':
    case 'x+':
      break;

    default:
      return false;
  }

  if ($fp = fopen($file_name, $mode));
  else
    return false;

  if (fputs($fp, $string));
  else
    return false;

  return fclose($fp);
}

/**
 * socket_set_option()'s
 *
 * iterate list of socket set options and return summation
 * NB, scalar lists key-paired values will first be cast
 * to their constant equivilant (if it has one) before being
 * assigned as a value
 * 
 * @param resource &$socket stock socket resource, see PHP sockets
 * @param integer $level see socket_set_option()
 * @param mixed $so_options string as scalar list or array of key-paired socket options, see socket_set_option()
 * @return ANDed return values from iterated socket_set_option() calls
*/ 
function sonic_socket_set_options(&$socket, $level, $so_options = SONIC_SOCK_OPTIONS)
{
  if (gettype($so_options) == 'integer')
  {
    $so_options = array($so_options);
  }
  else if (gettype($so_options) == 'array') {}
  else if (gettype($so_options) == 'string') //NB, only scalar set vals
  {
    $_so_options = array();
    $so_pts = preg_split('/,/', $so_options);
    foreach ($so_pts as $so_pt)
    {
      list($sopt, $soval) = preg_split('/=/', $so_pt, 2);
      if (($sopt = trim($sopt)) == NULL)
        continue;
      $_so_options[$sopt] = trim($soval);
    }
    $so_options = $_so_options;
    unset($_so_options);
  }
  else
    $so_options = array();

  $ret_all_ops = true;

  foreach ($so_options as $sopt => $soval)
  {
    if ($sopt == NULL)
      continue;

    //expects strings to be labels for constants
    if (gettype($sopt) == 'string')
    {
      if (strtoupper(substr($sopt, 0, 3)) == 'SO_');
      else
        $sopt = 'SO_'.$sopt;

      if (($sopt = @constant(strtoupper($sopt))) == NULL)
        continue;
    }

    if (isset($sopt) && isset($soval))
    {
      $ret_all_ops &= socket_set_option($socket, $level, $sopt, $soval);
    }
    else if (isset($soval))
    {
      $ret_all_ops &= socket_set_option($socket, $level, $soval);
    }
    else if (isset($sopt))
      $ret_all_ops &= socket_set_option($socket, $level, $sopt);
  }

  return $ret_all_ops;
}

/**
 * socket_write()
 *
 * socket_write() with write lock wait management
 *
 * @param resource $socket see PHP sockets
 * @param string $buffer see socket_write()
 * @param integer $length see socket_write()
 * @param string $lock_dir full path to scratch directory to be used 
 * @return integer see socket_write()
*/ 
function sonic_socket_write($socket, $buffer, $length, $lock_dir = SONIC_TMP_PATH)
{
  $socket_hash = '';

  if (socket_getsockname($socket, $local_ip, $local_port))
  {
    $local_port += 0;
    $socket_hash .= "$local_port.$local_ip";
  }

  if (socket_getpeername($socket, $remote_ip, $remote_port))
  {
    $remote_port += 0;
    $socket_hash .= "$remote_port.$remote_ip";
  }

  $socket_lock_file = $lock_dir.'/sonic-socket'.$socket_hash.'.lock';
//  $socket_lock_file = $lock_dir.'/sonic-socket'.md5($socket_hash).'lock';

  $time_now = time();
  $time_max = 5;
  while (file_exists($socket_lock_file))
  {
    $lock_time = filectime($socket_lock_file);

    if ($lock_time - $time_now > $time_max)
    {
      unlink($socket_lock_file);
      break;
    }

    // wait for 0.25 seconds
    usleep(250000);
  }

  touch($socket_lock_file);
  $out = socket_write($socket, $buffer, $length);
  unlink($socket_lock_file);

  return $out;
}

/**
 * buffer/string/packet/data checksum
 *
 * no and den...
 *
 * @link http://computer.howstuffworks.com/encryption7.htm
 * @param string $data raw data to check+sum
 * @param integer $bit_amt checksum of bit amount in size eg. 8, 16, 32, etc
 * @return string binary string of type pack()
*/ 
function sonic_checksum($data, $bit_amt = 32)
{ 
  $bit_amt += 0;

  if (($bit_amt % 8) == 0);
  else
    return false;

  $bit_mod = $bit_amt / 8;

  //2^8  - 1 =   256 - 1 per  8bits (1 byte) = 0xff
  //2^16 - 1 = 65536 - 1 per 16bits (2 byte) = 0xffff
  //2^32 - 1 = 4294967296 etc
  $bit_max = pow(2, $bit_amt) - 1;

  $bit = unpack('n*', $data);

  $cksum = array_sum($bit);

  //http://computer.howstuffworks.com/encryption7.htm
  //1. If the sum of the other bytes in the packet is 255 or less,
  //   then the checksum contains that exact value.
  //2. If the sum of the other bytes is more than 255, then the
  //   checksum is the remainder of the total value
  //   after it has been divided by 256.
  if (strlen($data) % 2)
  { 
    $flow = unpack('C*', $data[strlen($data) - 1]);
    $cksum += $flow[1];
  }

  $cksum  = ($cksum >> $bit_amt) + ($cksum & $bit_max);
  $cksum += ($cksum >> $bit_amt);
  $cksum =~ $cksum;

  return pack('n*', $cksum);
}

/**
 * sonic_http_env_engine() wrapper
 *
 * @param string $header_raw see sonic_http_env_engine()
 * @param array &$get_globals see sonic_http_env_engine()
 * @param array $default_index see sonic_http_env_engine()
 * @param string $tmp_dir see sonic_http_env_engine()
 * @return integer see sonic_http_env_engine()
*/ 
function sonic_http_env_get($header_raw, &$get_globals, $default_index = array('index.php', 'index.html'), $tmp_dir = SONIC_TMP_PATH)
{
  $get_globals = array();
  $opts['set_globals'] = false;
  $opts['get_globals'] = &$get_globals;
  $opts['default_index'] = $default_index;
  $opts['default_scheme'] = 'http';
  $opts['tmp_dir'] = $tmp_dir;

  return sonic_http_env_engine($header_raw, $opts);
}

/**
 * sonic_http_env_engine() wrapper
 *
 * @param string $header_raw see sonic_http_env_engine()
 * @param array $default_index see sonic_http_env_engine()
 * @param string $tmp_dir see sonic_http_env_engine()
 * @return interger see sonic_http_env_engine()
*/ 
function sonic_http_env($header_raw, $default_index = array('index.php', 'index.html'), $tmp_dir = SONIC_TMP_PATH)
{
  $opts['set_globals'] = true;
  $opts['default_index'] = $default_index;
  $opts['default_scheme'] = 'http';
  $opts['tmp_dir'] = $tmp_dir;

  return sonic_http_env_engine($header_raw, $opts);
}

//TODO: doc
function sonic_http_env_mini_engine($header_raw, $scheme = 'http')
{
  $header = sonic_header2array($header_raw);

  if (preg_match('/^\s*(?:GET|POST)\s+([^\s]*)\s+HTTP\/1\.[0-1]\s*$/im', $header_raw, $matches));
  else
    return false;

  $header_URL = $matches[1];
          
  $parse_URL = parse_url($header_URL);
          
  if ($_SERVER['QUERY_STRING'] = $parse_URL['query']);
  else
    unset($_SERVER['QUERY_STRING']);
        
  if ($_SERVER['SCRIPT_URL'] = $parse_URL['path']);
  else
    unset($_SERVER['SCRIPT_URL']);
          
  if ($tmp_SERVER['HTTP_HOST'] = $header['Host'])
  { 
    list($tmp_SERVER['SERVER_NAME'], $tmp_SERVER_PORT) = preg_split('/:/', $tmp_SERVER['HTTP_HOST']);
    
    if ($tmp_SERVER['SERVER_PORT']);
    else
      $tmp_SERVER['SERVER_PORT'] = $tmp_SERVER_PORT;
    
    if ($tmp_SERVER['SERVER_ADDR']);
    else
      $tmp_SERVER['SERVER_ADDR'] = $tmp_SERVER['SERVER_NAME'];
  }
  else
    unset($tmp_SERVER['HTTP_HOST']);
  
  if ($_SERVER['SCRIPT_URL'])
  {
    if ($scheme)
      $_SERVER['SCRIPT_URI'] = ''
                              .$scheme.'://'
                              .($tmp_SERVER['SERVER_NAME'] ? $tmp_SERVER['SERVER_NAME'] : $tmp_SERVER['SERVER_ADDR'])
                              .($tmp_SERVER['SERVER_PORT'] ? ':'.$tmp_SERVER['SERVER_PORT'] : '')
                              .$_SERVER['SCRIPT_URL'];
  }

  return true;
}

/**
 * setup PHP HTTP/CGI Environment
 *
 * will process given HTTP header key-paired tag formatted string (eg. usually a HTTP request)
 * and prepare and/or set PHP type superglobals for HTTP/CGI type environment
 *
 * where possible return values can be..
 *
 *1+:   OK
 *
 *0+:   not get|post aka unknown method
 *
 *-1:   bad get|post aka invalid syntax
 *
 *-2:   bad post
 *
 *-3:   script filename does not exist
 *
 *-4:   bad script aka invalid syntax
 *
 *-5:   no default index found
 *
 * @global array the PHP $_GET superglobal
 * @global array the PHP $_POST superglobal
 * @global array the PHP $_FILES superglobal
 * @global array the PHP $_COOKIE superglobal
 * @global array the PHP $_REQUEST superglobal
 * @global array the PHP $_SERVER superglobal
 * @param string $header_raw HTTP data (usually a HTTP request) containing GET, header key-paired values (eg. COOKIE) and POST 
 * @param array &$opts option flags, (boolean) set_globals, (boolean) get_globals, (array) default_index, string default_scheme
 * @return integer integer where greater than zero equates to true and zero or less false
*/ 
function sonic_http_env_engine($header_raw, &$opts)
{
  //this might not be the time saver we are looking for
  if (gettype($opts['header_parts']) == 'array')
    $header = $opts['header_parts'];
  else
    $header = false;

  if (isset($opts['set_globals']))
    $globals_set = $opts['set_globals'];
  else
    $globals_set = true;

  $globals_get = &$opts['get_globals'];
  if (isset($globals_get));
  else
    $globals_get = array();

  if (isset($opts['default_index']))
    $default_index = $opts['default_index'];
  else
    $default_index = array('index.php', 'index.html');

  if (isset($opts['default_scheme']))
    $default_scheme = $opts['default_scheme'];
  else
    $default_scheme = 'http';

  if (isset($opts['tmp_dir']))
    $tmp_dir = $opts['tmp_dir'];
  else
    $tmp_dir = SONIC_TMP_PATH;

  if ($globals_set) //superglobals much, is this even needed?
    global $_GET, $_POST, $_FILES, $_COOKIE, $_REQUEST, $_SERVER;

  //SERVER_ADMIN	webmaster@*.com	//parent
  //SERVER_PROTOCOL 	HTTP/1.1	//parent or default
  //GATEWAY_INTERFACE 	CGI/1.1		//parent 
  //SERVER_SIGNATURE 	no value	//parent
  //DOCUMENT_ROOT           /path/web	//parent or default
  //SERVER_SOFTWARE 	Apache/2.2.4	//parent

  $tmp_GET	= $_GET;
  $tmp_COOKIE	= $_COOKIE;
  $tmp_POST	= $_POST;
  $tmp_FILES	= $_FILES;
  $tmp_REQUEST	= $_REQUEST;
  $tmp_SERVER	= $_SERVER;

  if (preg_match('/^\s*(?:GET|POST)\s+([^\s]*)\s+HTTP\/1\.[0-1]\s*$/im', $header_raw, $matches));
  else
    return -1;

  $header_URL = $matches[1];

  $parse_URL = parse_url($header_URL);

  sonic_parse_str($parse_URL['query'], $tmp_REQUEST);
  $tmp_GET = $tmp_REQUEST; //php sets here

  if (preg_match('/^GET/im', $header_raw))
  {
    $method = 'get';
    $tmp_SERVER['REQUEST_METHOD'] = 'GET';
    //query in URL above
//    sonic_parse_str($parse_URL['query'], $tmp_GET);
//    $tmp_GET = $tmp_REQUEST; //assign. should be fine here
  }
  else if (preg_match('/^POST/im', $header_raw))
  {
    $method = 'post';
    $tmp_SERVER['REQUEST_METHOD'] = 'POST';

//TODO: multipart
    $simple_post = preg_split('/\r{0,1}\n\r{0,1}\n/', $header_raw);
    $simple_post_so = sizeof($simple_post);

    if ($simple_post_so == 2);
    else if ($simple_post_so > 2)
    {
      $simple_post = preg_split('/\r{0,1}\n\r{0,1}\n/', $header_raw, 2);

      if ($c_post_data = sonic_http_request_post($header_raw, $tmp_FILES, $tmp_dir));
      else
        return -2; //false; //$c_post_data;

      $simple_post = array($simple_post[0]);

      foreach ($c_post_data[1] as $pd_i => $pd_val)
      {
	$pd_name = $c_post_data[0][$pd_i];
        $simple_post[1] .= "&$pd_name=$pd_val";
      }

      $simple_post[1] = substr($simple_post[1], 1);
    }
    else
      return -2;

    unset($simple_post_so);

    if ($simple_post[1])
    {
      sonic_parse_str($simple_post[1], $tmp_POST);
      $tmp_REQUEST = array_merge($tmp_REQUEST, $tmp_POST);
      $parse_URL['query'] .= '#SONIC_WWW_POST#'.preg_replace('/\s+/', ' ', $simple_post[1]); //debug
    }
  }
  else
    return 0;

  //HTTPS check not implemented
  //not as if there is such animal as URL get/post compliance
  if ($parse_URL['scheme']) { }
  else if (preg_match('/^(.+):\/\//', $header_URL, $matches))
  {
    $parse_URL['scheme'] = $matches[1];
  }
  else
    $parse_URL['scheme'] = $default_scheme;
  unset($matches);

  if ($tmp_SERVER['DOCUMENT_ROOT']);
  else
    $tmp_SERVER['DOCUMENT_ROOT'] = '.';

  if ($tmp_SERVER['QUERY_STRING'] = $parse_URL['query']);
  else
    unset($tmp_SERVER['QUERY_STRING']);

  if ($tmp_SERVER['SCRIPT_URL'] = $parse_URL['path'])
  {
    //(at least) assign these before return for -3 AND -4 anyway
    $tmp_SERVER['SCRIPT_NAME'] = $tmp_SERVER['SCRIPT_URL'];
    $tmp_SERVER['SCRIPT_FILENAME'] = $tmp_SERVER['DOCUMENT_ROOT'].$tmp_SERVER['SCRIPT_URL'];
    $tmp_SERVER['REQUEST_URI'] = $tmp_SERVER['SCRIPT_URL']
			.($tmp_SERVER['QUERY_STRING'] ? '?'.$tmp_SERVER['QUERY_STRING'] : '');
    if ($tmp_SERVER['REQUEST_URI']);
    else
      unset($tmp_SERVER['REQUEST_URI']);
    $tmp_SERVER['PATH_TRANSLATED'] = $tmp_SERVER['SCRIPT_FILENAME'];
    $tmp_SERVER['PHP_SELF'] = $tmp_SERVER['SCRIPT_NAME']; //$tmp_SERVER["SCRIPT_URL"];

    //script filename check eg. ../
    if ($_sf_check = realpath($tmp_SERVER['SCRIPT_FILENAME'])) {}
    else if ($globals_set)
    {
      $globals_get['_SERVER'] = $tmp_SERVER;

//TODO: make this public option?
//OPT. A
//      $_SERVER = $tmp_SERVER;
//OPT. B
      sonic_http_env_server_var_reset();
      $_SERVER = array_merge($_SERVER, $tmp_SERVER);

      unset($tmp_SERVER);

      return -3;
    }
    else
    {
      $globals_get['_SERVER'] = $tmp_SERVER;

      return -3;
    }

    $_sf_check_p = str_replace(DIRECTORY_SEPARATOR, '\\'.DIRECTORY_SEPARATOR, $tmp_SERVER['DOCUMENT_ROOT']);
 
    if (preg_match("/^$_sf_check_p/i", $_sf_check))
      unset($_sf_check, $_sf_check_p);
    else //NB, prolly not safe to set globals?
    {
      $globals_get['_SERVER'] = $tmp_SERVER;

      return -4;
    }

//    $default_index = array("index.php", "index.html", "index.php3", "index.php4", "index.php5");
    if ($tmp_SERVER['SCRIPT_FILENAME'][strlen($tmp_SERVER['SCRIPT_FILENAME']) - 1] == '/')
    {
      $di_found = false;
      foreach ($default_index as $_di)
      {
        if (file_exists($tmp_SERVER['SCRIPT_FILENAME'].$_di))
        {
          $di_found = true;
          break;
        }
      }

      if ($di_found)
      {
        $tmp_SERVER['SCRIPT_FILENAME'] = $tmp_SERVER['SCRIPT_FILENAME'].$_di;
        $tmp_SERVER['SCRIPT_NAME'] = $tmp_SERVER['SCRIPT_NAME'].$_di;
      }
      else if ($globals_set)
      {
//TODO: make this public option?
//OPT. A
//        $_SERVER = $tmp_SERVER;
//OPT. B
        sonic_http_env_server_var_reset();
        $_SERVER = array_merge($_SERVER, $tmp_SERVER);

        unset($tmp_SERVER);

        return -5;
      }
    }
  }
  else
    unset($tmp_SERVER['SCRIPT_URL']);

  if ($header);
  else
    $header = sonic_header2array($header_raw);

  if ($tmp_SERVER['SERVER_PROTOCOL']);
  else
    $tmp_SERVER['SERVER_PROTOCOL'] = 'HTTP/1.0';

  if ($tmp_COOKIE = $header['Cookie'])
  {
    sonic_parse_str($tmp_COOKIE, $tmp_COOKIE);
    $tmp_REQUEST = array_merge($tmp_REQUEST, $tmp_COOKIE);
  }
  else
    unset($tmp_COOKIE);

  if ($tmp_SERVER['PHPRC'] = ini_get('include_path'));
  else
    unset($tmp_SERVER['PHPRC']);

  if ($tmp_SERVER['CONTENT_LENGTH'] = $header['Content-Length']);
  else
    unset($tmp_SERVER['CONTENT_LENGTH']);

  if ($tmp_SERVER['CONTENT_TYPE'] = $header['Content-Type']);
  else
    unset($tmp_SERVER['CONTENT_TYPE']);

  if ($tmp_SERVER['HTTPS'] = (($parse_URL['scheme'] == 'https') ? 'on' : NULL));
  else
    unset($tmp_SERVER['HTTPS']);

  if ($tmp_SERVER["HTTP_USER_AGENT"] = $header["User-Agent"]);
  else
    unset($tmp_SERVER["HTTP_USER_AGENT"]);

  if ($tmp_SERVER["HTTP_ACCEPT"] = $header["Accept"]);
  else
    unset($tmp_SERVER["HTTP_ACCEPT"]);

  if ($tmp_SERVER["HTTP_ACCEPT_LANGUAGE"] = $header["Accept-Language"]);
  else
    unset($tmp_SERVER["HTTP_ACCEPT_LANGUAGE"]);

  if ($tmp_SERVER["HTTP_ACCEPT_ENCODING"] = $header["Accept-Encoding"]);
  else
    unset($tmp_SERVER["HTTP_ACCEPT_ENCODING"]);

  if ($tmp_SERVER["HTTP_ACCEPT_CHARSET"] = $header["Accept-Charset"]);
  else
    unset($tmp_SERVER["HTTP_ACCEPT_CHARSET"]);

  if ($tmp_SERVER["HTTP_CACHE_CONTROL"] = $header["Cache-Control"]);
  else
    unset($tmp_SERVER["HTTP_CACHE_CONTROL"]);

  if ($tmp_SERVER["HTTP_CONNECTION"] = $header["Connection"]);
  else
    unset($tmp_SERVER["HTTP_CONNECTION"]);

  if ($tmp_SERVER['HTTP_HOST'] = $header['Host'])
  {
    list($tmp_SERVER['SERVER_NAME'], $tmp_SERVER_PORT) = preg_split('/:/', $tmp_SERVER['HTTP_HOST']);
//NB, might be virtual host entry and not loaded interface
//    if ($tmp_SERVER['SERVER_PORT']);
//    else
      $tmp_SERVER['SERVER_PORT'] = $tmp_SERVER_PORT;
//NB, see above
//    if ($tmp_SERVER['SERVER_ADDR']);
//    else
      $tmp_SERVER['SERVER_ADDR'] = $tmp_SERVER['SERVER_NAME'];
  }
  else
    unset($tmp_SERVER['HTTP_HOST']);

  if ($tmp_SERVER['SCRIPT_URL'])
  {
      $tmp_SERVER['SCRIPT_URI'] = ''
		.$parse_URL['scheme'].'://'
		.($tmp_SERVER['SERVER_NAME'] ? $tmp_SERVER['SERVER_NAME'] : $tmp_SERVER['SERVER_ADDR'])
		.($tmp_SERVER['SERVER_PORT'] ? ':'.$tmp_SERVER['SERVER_PORT'] : '')
		.$tmp_SERVER['SCRIPT_URL']
		.'';
  }

  if ($tmp_SERVER['HTTP_KEEP_ALIVE'] = $header['Keep-Alive']);
  else
    unset($tmp_SERVER['HTTP_KEEP_ALIVE']);

  if ($tmp_SERVER['HTTP_REFERER'] = $header['Referer']);
  else
    unset($tmp_SERVER['HTTP_REFERER']);

  if ($globals_set)
  {
    $_GET	= $tmp_GET;
    $_COOKIE	= $tmp_COOKIE;
    $_POST	= $tmp_POST;
    $_FILES	= $tmp_FILES;
    $_REQUEST	= $tmp_REQUEST;

//TODO: make this public option?
//OPT. A
//    $_SERVER	= $tmp_SERVER;
//OPT. B
    sonic_http_env_server_var_reset();
    $_SERVER	= array_merge($_SERVER, $tmp_SERVER);
  }

  if (gettype($globals_get) == 'array');
  else
    $globals_get = array();

  $globals_get['_GET']		= $tmp_GET;
  $globals_get['_COOKIE']	= $tmp_COOKIE;
  $globals_get['_POST'] 	= $tmp_POST;
  $globals_get['_FILES'] 	= $tmp_FILES;
  $globals_get['_REQUEST'] 	= $tmp_REQUEST;
  $globals_get['_SERVER'] 	= $tmp_SERVER;

  return 1;
}

//TODO: func. info
function sonic_http_env_server_var_reset()
{
    unset($_SERVER['DOCUMENT_ROOT'],
          $_SERVER['QUERY_STRING'],
          $_SERVER['SCRIPT_URL'],
          $_SERVER['SCRIPT_NAME'],
          $_SERVER['SCRIPT_FILENAME'],
          $_SERVER['REQUEST_URI'],
          $_SERVER['PATH_TRANSLATED'],
          $_SERVER['PHP_SELF'],
          $_SERVER['SERVER_PROTOCOL'],
          $_SERVER['PHPRC'],
          $_SERVER['CONTENT_LENGTH'],
          $_SERVER['CONTENT_TYPE'],
          $_SERVER['HTTPS'],
          $_SERVER['HTTP_USER_AGENT'],
          $_SERVER['HTTP_ACCEPT'],
          $_SERVER['HTTP_ACCEPT_LANGUAGE'],
          $_SERVER['HTTP_ACCEPT_ENCODING'],
          $_SERVER['HTTP_ACCEPT_CHARSET'],
          $_SERVER['HTTP_CACHE_CONTROL'],
          $_SERVER['HTTP_CONNECTION'],
          $_SERVER['HTTP_HOST'],
          $_SERVER['SERVER_PORT'],
          $_SERVER['SERVER_ADDR'],
	  $_SERVER['SCRIPT_URI'],
          $_SERVER['HTTP_KEEP_ALIVE'],
          $_SERVER['HTTP_REFERER']);
}

/**
 * setup $_FILES from HTTP request
 *
 * process HTTP request and setup PHP superglobal $_FILES type array
 *
 * @param string $http_request HTTP request data to process post data from
 * @param &$_FILES_PHP array of PHP type superglobal $_FILES
 * @return array array of post data key and value arrays
*/ 
function sonic_http_request_post($http_request, &$_FILES_PHP, $tmp_dir = SONIC_TMP_PATH)
{
  //split request header into array of vars for curl
  $hhr_split_p = '/\r{0,1}\n\r{0,1}\n/';
  //header + get/post AND post data
  list($hhr_header, $hhr_body) = preg_split($hhr_split_p, $http_request, 2);

  //header in array (- get/post)
  $hhr_header_a = sonic_header2array($hhr_header);

  //cycle body as post data
  $matches = array();
  $hhr_blocks_a = array();
  $_FILES_PHP = array();
  $c_post_data_k = $c_post_data_v = array();

  //Content-Type: multipart/form-data; boundary=DELIM
  if (preg_match('/\s*multipart\s*[^\s]\s*form-data;\s+boundary\s*=([^\r\n]+)$/i', $hhr_header_a['Content-Type'], $matches))
  {
//WARN
//    $hhr_blocks_a = preg_split("/^".$matches[1]."\r{0,1}\n/m", $hhr_body); //should be this
// but curl is fracked sending boundary two hypens short in header
// ASWELL AS two hypens extra at tail
    $hhr_blocks_a = preg_split('/^[\s-]*'.$matches[1].'[\s-]*\r{0,1}\n/m', $hhr_body);
  }
  //Content-Type: application/x-www-form-urlencoded (DEFAULT NO CHECK)
  else if (preg_match('/\s*application\s*[^\s]\s*x-www-form-/i', $hhr_header_a['Content-Type'], $matches))
  {
    sonic_parse_str_kv(preg_replace('/\r{0,1}\n$/', '', $hhr_body), $c_post_data_k, $c_post_data_v);
  }
  //Boundary: DELIM
  else if ($hhr_header_a['Boundary'] != '')
  {
    $hhr_blocks_a = preg_split('/^'.$hhr_header_a['Boundary'].'\r{1,1}\n/m', $hhr_body);
  }
  else
  {
//    echo "post not handled yet\n";
    return false;
  }

  if (sizeof($hhr_blocks_a) <= 0); //no post data
  else
  { 
    foreach ($hhr_blocks_a as $hhr_block)
    {
      if ($hhr_block == '')
        continue;

      //block header
      list($block_header, $block_body) = preg_split($hhr_split_p, $hhr_block, 2);

      $block_body = preg_replace('/\r{0,1}\n$/', '', $block_body);

      $block_headers = sonic_header2array($block_header);

      //relevant header check for file
      //Content-Type: multipart/form-data; boundary=DELIM
      if ($block_headers['Content-Disposition']
       && preg_match('/form-data[^a-z0-9]*/i', $block_headers['Content-Disposition'])
         )
      {
	$matches = array();
	//Content-Disposition: form-data; name="post_var"
	if (preg_match('/\s+name\s*=\s*[\'"]{0,1}([^\r\n\'"]+)[\'"]{0,1}/i', $block_headers['Content-Disposition'], $matches))
	{
	  $c_post_data_k[] = $matches[1];
	  $c_post_data_v[] = $block_body;

	  $matches_fn = array();
	  if (preg_match('/\s+filename\s*=\s*[\'"]{0,1}([^\r\n\'"]+)[\'"]{0,1}/i', $block_headers['Content-Disposition'], $matches_fn))
	  {
//http://au2.php.net/manual/en/features.file-upload.errors.php
//TODO: compliance much?
	    $c_post_v_n = $matches_fn[1];
	    $c_post_f_n = $tmp_dir.DIRECTORY_SEPARATOR.'sonic_'.md5($c_post_v_n);
	    $c_post_f_upload_error = UPLOAD_ERR_OK;
	    if ($c_post_f_n_check = (strlen($block_body) > 0))
	    {
	      if ($c_post_f_tmp_dir_check = is_dir($tmp_dir));
	      else
		$c_post_f_upload_error = UPLOAD_ERR_NO_TMP_DIR;

	      if ($c_post_f_tmp_dir_check)
	      {
		if ($c_post_f_h = @fopen($c_post_f_n, 'wb'))
		{
		  fputs($c_post_f_h, $block_body);
		  fclose($c_post_f_h);
//		  $c_post_data_k[] = $matches[1];
//		  $c_post_data_v[] = '@'.$c_post_f_n; //debug

//WARN+TODO: apache reports +1 check block_body if missing last char?
		  $c_post_f_so = filesize($c_post_f_n);

		  $ig_so = ini_get('upload_max_filesize');
		  if (gettype($ig_so) == 'integer') {}
		  else if (gettype($ig_so) == 'string')
		  {
		    $ig_so = sonic_byte_longhand($ig_so);
		  }
		  else
		    $ig_so = 0;

		  if ($c_post_f_so > $ig_so && $ig_so > 0)
		  {
		    $c_post_f_upload_error = UPLOAD_ERR_INI_SIZE;
		  }
		  else if ($c_post_f_so > MAX_FILE_SIZE && MAX_FILE_SIZE > 0)
		  {
		    $c_post_f_upload_error = UPLOAD_ERR_FORM_SIZE;
		  }
		}
		else
		{
		  $c_post_f_upload_error = UPLOAD_ERR_CANT_WRITE;
		  $c_post_f_so = 0;
		  unset($c_post_f_h);
		}
	      }
	    }
	    else
	    {
	      $c_post_f_upload_error = UPLOAD_ERR_NO_FILE;
	      $c_post_f_so = 0;
	      unset($c_post_f_h);
	    }
//TODO: NFI -> UPLOAD_ERR_EXTENSION

	    $c_post_f_data = array();
	    $c_post_f_data['name'] = $c_post_v_n;
	    $c_post_f_data['type'] = sonic_mimetype($c_post_v_n);
	    $c_post_f_data['tmp_name'] = $c_post_f_n;
	    $c_post_f_data['error'] = $c_post_f_upload_error;
	    $c_post_f_data['size'] = $c_post_f_so;

	    $_FILES_PHP["$matches[1]"] = $c_post_f_data;
	  }
	  unset($matches_fn);
	  unset(	$c_post_f_data,
			$c_post_v_n,
			$c_post_f_n,
			$c_post_f_upload_error,
			$c_post_f_so,
			$c_post_f_n_check,
			$c_post_f_tmp_dir_check);
	}
	else
//TODO: check for more sub-meta methods for CD
	  continue;
      }
      //Boundary: DELIM
      else if ($block_headers['Filename'] && $block_headers['Name'])
      {
	$file = $tmp_dir.DIRECTORY_SEPARATOR.$block_headers['Filename'];

	if ($fp = fopen($file, 'wb'));
	else
	  break;

	fputs($fp, $block_body);
	fclose($fp);

	$c_post_data_k[] = $block_headers['Name'];
	$c_post_data_v[] = '@'.$file;
      }
      else if ($block_headers['Name'])
      {
	$c_post_data_k[] = $block_headers['Name'];
	$c_post_data_v[] = $block_body;
      }
      //todo:TODO
      else
        break;
    }
  }

  return array($c_post_data_k, $c_post_data_v);
}

/**
 * mimetype
 *
 * will attempt to use any inbuilt mime_content_type() function first before
 * loading its own
 *
 * @param string $file location to file requiring mime definition lookup (via filename extension)
 * @return string mime type for $file or NULL on any failure
*/ 
function sonic_mimetype($file)
{
  $_fe = explode('.', $file);
  $_fes = sizeof($_fe);

  if ($_fes > 0);
  else
    return;

  $ext = array_pop($_fe);
  if ($_fes == 1)
    $ext = basename($ext);

  unset($_fe, $_fes);

  if (function_exists('mime_content_type'));
  else
  { 
    function mime_content_type($file)
    { 
      $fi = finfo_open(FILEINFO_MIME);
      $mimetype = finfo_file($fi, $file);
      finfo_close($fi);
      return $mimetype;
    }
  }

  if (file_exists($file));
  else
    return;

  if ($out_m = mime_content_type($file))
    return $out_m;
  else
//    return 'application/octet-stream'; //run dont ask
    return 'application/x-octet-stream';
}

/**
 * HTTP header check
 *
 * will make sure that essential header parts are of standard
 * word casing to prevent double hashing of same header via
 * mismatch case indexing
 *
 * @param array $http_headers_a HTTP header array, see sonic_header2array()
 * @return array array of HTTP headers of type sonic_header2array()
*/ 
function sonic_http_header_check($http_headers_a = array())
{
  if (gettype($http_headers_a) == 'array');
  else
    return array();

  //TODO: keep synced with sonic_http_response(), ie;
  //only really need to check possible sonic/internal 
  //header subsitutes are correct ie. currently;
//	  Content-Type
//	  Connection
//	  Content-Length
//	  Transfer-Encoding
//	  Date
  //why?
  //dont want to WWW-Authenticate->Www-Authenticate
  //+ leave user headers alone if possible :p
  foreach ($http_headers_a as $h_k => $h_v)
  {
    switch (strtolower($h_k))
    {
      case 'content-type':
      case 'connection':
      case 'content-length':
      case 'transfer-encoding':
      case 'date':
        unset($http_headers_a[$h_k]);
//        $h_k = ucwords(strtolower($h_k));
        $h_k = join('-', array_map('ucwords', explode('-', strtolower($h_k))));
        $http_headers_a[$h_k] = $h_v;
      break;

      default: 
    }
  }

  return $http_headers_a;
}

/**
 * form HTTP response from parts
 *
 * will merge the supplied parts given and fill-in any gaps
 * that are missing to form a single compliant reponse
 *
 * @param integer $http_response_code response code to merge with response, eg. 200
 * @param string $http_body HTTP body data to merge with response
 * @param array $http_headers_a array of HTTP headers of type sonic_header2array() to merge with response
 * @return string HTTP compliant response
*/ 
function sonic_http_response($http_response_code, $http_body, $http_headers_a = array())
{
  if ($_mime_type = $http_headers_a['Content-Type']) {}
  else if ($_mime_type = sonic_mimetype($_SERVER['SCRIPT_FILENAME'])) {}
  else
    $_mime_type = 'text/plain';

  $http_response_code += 0;
  switch ($http_response_code)
  {
    case 100:
      $_header_head = 'Continue';
    break;
    case 101:
      $_header_head = 'Switching Protocols';
    break;

    case 200:
      $_header_head = 'OK';
    break;
    case 201:
      $_header_head = 'Created';
    break;
    case 202:
      $_header_head = 'Accepted';
    break;
    case 203:
      $_header_head = 'Non-Authoritative Information';
    break;
    case 204:
      $_header_head = 'No Content';
    break;
    case 205:
      $_header_head = 'Reset Content';
    break;
    case 206:
      $_header_head = 'Partial Content';
    break;

    case 300:
      $_header_head = 'Multiple Choices';
    break;
    case 301:
      $_header_head = 'Moved Permanently';
    break;
    case 302:
      $_header_head = 'Found';
    break;
    case 303:
      $_header_head = 'See Other';
    break;
    case 304:
      $_header_head = 'Not Modified';
    break;
    case 305:
      $_header_head = 'Use Proxy';
    break;
    //306 reserved
    case 307:
      $_header_head = 'Temporary Redirect';
    break;

    case 400:
      $_header_head = 'Bad Request';
    break;
    case 401:
      $_header_head = 'Authorization Required'; //'Unauthorized';
    break;
    case 402:
      $_header_head = 'Payment Required';
    break;
    case 403:
      $_header_head = 'Forbidden';
    break;
    case 404:
      $_header_head = 'Not Found';
    break;
    case 405:
      $_header_head = 'Method Not Allowed';
    break;
    case 406:
      $_header_head = 'Not Acceptable';
    break;
    case 407:
      $_header_head = 'Proxy Authentication Required';
    break;
    case 408:
      $_header_head = 'Request Timeout';
    break;
    case 409:
      $_header_head = 'Conflict';
    break;
    case 410:
      $_header_head = 'Gone';
    break;
    case 411:
      $_header_head = 'Length Required';
    break;
    case 412:
      $_header_head = 'Precondition Failed';
    break;
    case 413:
      $_header_head = 'Request Entity Too Large';
    break;
    case 414:
      $_header_head = 'Request-URI Too Long';
    break;
    case 415:
      $_header_head = 'Unsupported Media Type';
    break;
    case 416:
      $_header_head = 'Requested Range Not Satisfiable';
    break;
    case 417:
      $_header_head = 'Expectation Failed';
    break;

    case 500:
      $_header_head = 'Internal Server Error';
    break;
    case 501:
      $_header_head = 'Not Implemented';
    break;
    case 502:
      $_header_head = 'Bad Gateway';
    break;
    case 503:
      $_header_head = 'Service Unavailable';
    break;
    case 504:
      $_header_head = 'Gateway Timeout';
    break;
    case 505:
      $_header_head = 'HTTP Version Not Supported';
    break;

    default:
//      $http_response_code = 200;
//      $_header_head = 'OK';
      $http_response_code = 501;
      $_header_head = 'Not Implemented';
    break;
  }

  $_time = time();

  if ($_conn = $http_headers_a['Connection']);
  else
    $_conn = ((preg_match('/(?:text|html|plain)/', $_mime_type)) ?  'close' : 'Keep-Alive');

  if ($_len = $http_headers_a['Content-Length']);
  else
    $_len = strlen($http_body);

  if ($_date = $http_headers_a['Date']);
  else
    $_date = gmdate('D, d M Y H:i:s', $_time).' GMT'; //apache shows local GMT not GMT+0

  unset($http_headers_a['Content-Type']);
  unset($http_headers_a['Connection']);
  unset($http_headers_a['Content-Length']);
  unset($http_headers_a['Date']);

  $http_header_rest = '';
  foreach ($http_headers_a as $h_key => $h_val)
  {
    if (preg_match('/:/', $h_key))
      $http_header_rest .= "$h_key$h_val\n";
    else
      $http_header_rest .= "$h_key: $h_val\n";
  }

  $out = '' 
         .'HTTP/1.x '.$http_response_code.' '.$_header_head."\n"
         .'Date: '.$_date."\n"
//         .((preg_match('/(?:text|html|plain)/', $_mime_type)) ?  '' : "Accept-Ranges: bytes\n")
         .$http_header_rest
//         .'';

//  $out ='' 
//A server MUST NOT send transfer-codings to an HTTP/1.0 client. - bad luck hey
         .((!preg_match('/identity/i', $http_headers_a['Transfer-Encoding']) && 
            !preg_match('/any/i', $http_headers_a['Content-Length'])
           )
          ?
           'Content-Length: '.$_len."\n"
          :
           '' 
          )
//         .'';

//  $out ='' 
         .'Connection: '.$_conn."\n"
         .'Content-Type: '.$_mime_type."\n"
         ."\n"
         .$http_body
         .'';

  return $out;
}

//DEV FUNKS
//function sonic_ext_check() {} //moved because of declare tick
//eqsf -> escaped or quoted space fix
/**
 * argv escaped/quoted string fix
 *
 * will process standard argv type array and concat/implode where needed
 * eg. escaped characters or quoted strings
 * 
 * @param array $argv runtime style arguments with escaped quotes and spaces
 * @return array array of quoted space corrected runtime style arguments
*/ 
function sonic_args_qsf($argv)
{
  for ($i = 1, $x = sizeof($argv); $i < $x; $i++)
  { 
    $arg = $argv[$i];

    if (preg_match('/\\\$/', $arg))
    {
      array_splice($argv, $i, 2, substr($argv[$i], 0, -1).' '.$argv[$i+1]);
      $x = sizeof($argv);
      $i--; //incase current shift ends in slash aswell
    }
   else if (preg_match('/=\'.*[^\']$/', $arg) OR preg_match('/=".*[^"]$/', $arg))
   {
     array_splice($argv, $i, 2, $argv[$i].' '.$argv[$i+1]);
     $x = sizeof($argv);
     $i--; //incase current shift contains quote
   }
   else if (preg_match('/([^=]+)=\'(.*)\'$/', $arg, $match) OR preg_match('/([^=]+)="(.*)"$/', $arg, $match))
   {
     $argv[$i] = $match[1].'='.$match[2];
   }
  }

  return $argv;
}

/**
 * get argv 
 *
 * custom argv manipulator function 
 *
 * @param array $argv array of stock standard argv params
 * @return array array of re-arranged/styled runtime arguments
*/ 
function sonic_args($argv)
{
  $argv = sonic_args_qsf($argv);

  $args = array();
  for ($i = 1, $x = sizeof($argv); $i < $x; $i++)
  {
    $arg = $argv[$i];

    //i => -X
    if (preg_match('/^-([^-\s\'"])$/', $arg, $match))
    {
      // i => -X
      if (!$argv[$i+1] OR preg_match('/^-/', $argv[$i+1]))
        $args[$match[1]] = true;
      // i => -X word (i+1)
      else
      {
        $i++;
        $args[$match[1]] = $argv[$i];
      }
    }
    //i => -XX => -X -X
    else if (preg_match('/^-([^-\s\'"])([^-\s\'"]+)$/', $arg, $match))
    {
      $args[$match[1]] = true;
      array_splice($argv, $i, 1, array($match[2], '-'.$match[2]));
      $x = sizeof($argv);
    }
    else if (preg_match('/^--([^=]+)$/', $arg, $match))
    {
      $arg = $match[1];
      $i++;

      if ($argv[$i] == '=')
        $i++;
      if (preg_match('/^={1,1}(.+)$/', $argv[$i], $match))
        $args[$arg] = $match[1];
      else
      {
        $args[$arg] = true;
        $i--;    //rewind
        continue;//reminders sake
      }

    }
    else if (preg_match('/^--([^=]+)=(.*)$/', $arg, $match))
    {
      $args[$match[1]] = $match[2];
    }
    else
      array_push($args, $arg);
  }

  /*
  while (list($flag, $value) = each($args))
    echo "arg[$flag] $value\n";
  */

  return $args;
}

//TODO: func. info
function sonic_array2str($array_res, $ue = false)
{
  $out = '';
  $null_str = 'NULL';

  foreach ($array_res as $k => $v)
  {
    if (is_array($v))
      $v = '^^array^'.sonic_array2str($v);
    else
      $v = (strlen($v) > 0) ? $v : $null_str;

    $k = urlencode($k);
    $v = urlencode($v);

    $out .= "$k|$v||";
  }

  if ($ue)
    return urlencode(substr($out, 0, -2));
  else
    return substr($out, 0, -2);
}

//TODO: func. info
function sonic_str2array($string_res, $ud = false)
{ 
  $out = array();
  $null_str = urlencode('NULL');

  if ($ud)
    $string_res = urldecode($string_res);
  
  $_out = explode('||', $string_res);
  
  foreach ($_out as $_v)
  { 
    list($k, $v) = explode('|', $_v);                                             
                                                                                  
    $k = urldecode($k);
    $v = urldecode($v);

    if ($v == $null_str)
      $out[$k] = NULL;
    else
    { 
      if (substr($v, 0, 8) == '^^array^')
        $v = sonic_str2array(substr($v, 8));

      $out[$k] = $v;
    }
  }

  return $out;
}

/**
 * HTTP string to array conversion (header tag filter)
 *
 * filters all HTTP type header key-paired tags from any given string
 * into array
 *
 * @param string $header string to ciphen HTTP headers from
 * @param string $return_type 'array', 'mixed' or default (clobber) values
 * @return array array of header tags with their respective values
*/
function sonic_header2array($header, $return_type = NULL)
{
  if (preg_match_all('/^(.+?): (.*?)\s*$/m', $header, $raw_header_tags)) {}
  else if (preg_match_all('/^(.+?):(.*)$/m', $header, $raw_header_tags)) {}

  switch ($return_type)
  {
    case 'array':
      for ($i = 0, $x = sizeof($raw_header_tags[1]); $i < $x; $i++)
      {
	if ($header_tags[$raw_header_tags[1][$i]] === NULL)
	  $header_tags[$raw_header_tags[1][$i]] = array($raw_header_tags[2][$i]);
	else
	  $header_tags[$raw_header_tags[1][$i]][] = $raw_header_tags[2][$i];
      }
    break;

    case 'mixed':
      for ($i = 0, $x = sizeof($raw_header_tags[1]); $i < $x; $i++)
      { 
	if ($header_tags[$raw_header_tags[1][$i]] === NULL)
	  $header_tags[$raw_header_tags[1][$i]] = $raw_header_tags[2][$i];
	else
	{
	  if (gettype($header_tags[$raw_header_tags[1][$i]]) == 'string')
	    $header_tags[$raw_header_tags[1][$i]] = array($header_tags[$raw_header_tags[1][$i]]);

	  $header_tags[$raw_header_tags[1][$i]][] = $raw_header_tags[2][$i];
	}
      }
    break;

    default:
      for ($i = 0, $x = sizeof($raw_header_tags[1]); $i < $x; $i++)
	$header_tags[$raw_header_tags[1][$i]] = $raw_header_tags[2][$i];
  }

  return $header_tags;
}

/**
 * date time 
 *
 * current time by default or supplied date format
 *
 * @param string $format date format string 
 * @return string time by date formatted string
*/ 
function sonic_datef($format = SONIC_DATEF) { return date($format, time()); }

/**
 * byte, string to integer conversion
 *
 * reverse map string representation into actual byte amount
 *
 * @param string $size byte size eg. 101K
 * @return integer byte size as an integer
*/ 
function sonic_byte_longhand($size)
{
  if (gettype($size) == 'string');
  else
    return $size + 0;

  $size_i = strlen($size) - 1;
  $bytes = substr($size, 0, $size_i) + 0;
  $short_hand = substr($size, $size_i);

  switch (strtolower($short_hand))
  {
    case 'k':
      $bytes *= 1024;
    break;
    case 'm':
      $bytes *= 1048576;
    break;
    case 'g':
      $bytes *= 1073741824;
    break;
  }

  return $bytes;
}

//TODO: info.
function sonic_srand_seeder()
{ 
  list($usec, $sec) = explode(' ', microtime());
  return (float) $sec + ((float) $usec * 100000);
}

//TODO: info
function sonic_shm_memheadersize()
{
  return 44;
//  return (PHP_INT_SIZE * 4) + 8;
}

//TODO: info
function sonic_shm_memsize($obj, $header = false)
{
  if ($header)
    $out = (sonic_shm_memheadersize() + strlen(serialize($obj)));
  else
    $out = strlen(serialize($obj));

//    $out = (int) ((((strlen(serialize($obj)))+ (4 * PHP_INT_SIZE)) /4 ) * 4 ) + 4; 

  return $out;
}

//NB, converts spaces and dots into underscores
//    parse_str($str, $arr);
function sonic_parse_str($str, &$arr, $method = NULL)
{
  $arr = array();
  $pairs = explode('&', $str);

  switch ($method)
  {
    case 'no_urldecode':
      foreach ($pairs as $pair)
      {
	list($k, $v) = explode('=', $pair);
	$arr[$k] = $v;
      }
    break;

    case 'urldecode_no_plus':
      foreach ($pairs as $pair)
      {
	list ($k, $v) = explode('=', $pair);

	$k = str_replace('+', '%2B', $k);
	$v = str_replace('+', '%2B', $v);

	$k = urldecode($k);
	$v = urldecode($v);

	$arr[$k] = $v;
      }
    break;

    default:
      foreach ($pairs as $pair)
      {
	list($k, $v) = array_map('urldecode', explode('=', $pair));
	$arr[$k] = $v;
      }
    break;
  }
}

function sonic_parse_str_kv($str, &$arr_k, &$arr_v, $method = NULL)
{
  $arr = array();
  $pairs = explode('&', $str);

  switch ($method)
  {
    case 'no_urldecode':
      foreach ($pairs as $pair)
	list($arr_k[], $arr_v[]) = explode('=', $pair);
    break;

    case 'urldecode_no_plus':
      foreach ($pairs as $pair)
      {
	list ($k, $v) = explode('=', $pair);

	$arr_k[] = urldecode(str_replace('+', '%2B', $k));
	$arr_v[] = urldecode(str_replace('+', '%2B', $v));
      }
    break;

    default:
      foreach ($pairs as $pair)
	list($arr_k[], $arr_v[]) = array_map('urldecode', explode('=', $pair));
    break;
  }
}


?>
