<?
/* --------------------------------------------------------------------
   Copyright (c) 2008-2012 Daniel Pedemont (dan@pedemont.net)
   -------------------------------------------------------------------- */
/**
 * stream exec
 *
 * drop shell executer
 *
 * @link http://dev.pedemont.com/sonic/
 * @link http://dev.pedemont.com/sonic/downloads/src/lib/streams/sonic.stream.exec.php
 * @package sonic_stream
 * @version 0.1.2
 * @abstract shell execution
*/

define('SONIC_EXEC_VERSION', '0.1.2+');

//NB, function names in rotate mode have no protection,
//    so dont use generic function names for clash safety

class SONIC_STREAM_exec extends SONIC_stream
{
  /**
   * popen resource handle
   *
   * @see popen()
   * @var resource
  */
  var $exec_handle = NULL;

  /**
   * signal termination handler
   *
   * @param integer $sig posix process signal
  */
  function _sonic_stream_exec_handler($sig = NULL)
  {
//echo "_sonic_stream_exec_handler()\n";
    //none of this is really needed, PHP will clean up everything
    //but useful for error messages with server reset in IPC mode
    //
    //in thread mode, sonic_thread_handler() will close
    //socket on last instance/thread anyway
    switch ($sig)
    { 
      case SIGTERM:
      case SIGQUIT:
      case SIGINT:
        if (($this->pid == 0 && $this->pid_child == 0) OR ($this->pid == $this->pid_child)) //rotate mode
        { 
          sonic_daemon_thread_shutdown($sig);
          $this->rotate_shutdown = true;
        }
        else
        { 
          $this->sonic_stream_exec_stop();
          $this->sonic_stream_buffer();
          exit();
        }
      break;

      default:
        die("exec says WTF");
    }
  }

  /**
   * main setup
   *
   * load signal handers then return
   *
   * @see SONIC_STREAM_exec::sonic_stream_main()
  */
  function sonic_stream_exec_start()
  {
    //TODO, need all of these?
    if ($this->pcntl_loaded)
    {
      pcntl_signal(SIGTERM, array(&$this, "_sonic_stream_exec_handler"));
      pcntl_signal(SIGQUIT, array(&$this, "_sonic_stream_exec_handler"));
      pcntl_signal(SIGINT, array(&$this, "_sonic_stream_exec_handler"));
    }

//    var_dump($this->config);
  }

  /**
   * main shutdown/finish
   *
   * closes popen resource variable and echos trace
   *
   * @see SONIC_STREAM_exec::sonic_stream_main()
   * @see SONIC_STREAM_exec::sonic_stream_exec_prelog()
  */
  function sonic_stream_exec_stop()
  {
    if ($this->verbose > 0)
    {
      $stream_config = $this->config["stream_config"];

      if ($msg_close = $stream_config["example_msg3"])
        $msg_close .= "\n";
      else
        $msg_close = "connection finished\n";

      echo '= '.$this->sonic_stream_exec_prelog().' ';
      echo $msg_close;
    }

    if ($this->exec_handle)
      pclose($this->exec_handle);
  }

  /**
   * stream main
   *
   * @see SONIC_stream::sonic_stream_main()
   * @see SONIC_STREAM_exec::sonic_stream_exec_prelog()
  */
  function sonic_stream_main()
  {
    $this->sonic_stream_exec_start();

    $stream_config = $this->config["stream_config"];

//var_dump($this->config);

    if ($exec = $stream_config['exec']);
    else
      die('! '.$this->sonic_stream_exec_prelog().' Error: main(): no \'exec\' option set'."\n");

    if ($this->verbose > 0)
    {
      echo '= '.$this->sonic_stream_exec_prelog().' ';
      echo 'Notice: executing \''.$exec.'\''."\n";
    }

    $this->idle = false;

    if ($this->exec_handle = popen($exec, 'r'));
    else
      die('! '.$this->sonic_stream_exec_prelog().' Error: main(): execute failed'."\n");

    $eh = &$this->exec_handle;

//TODO: have this write output to a file or summat
    while (feof($eh) == false)
    {
      $this->sonic_stream_buffer();

      $buffer = fgets($eh);

//TODO: double check \n should be here
      echo "$buffer\n";
      ob_flush();

      flush();

      if ($this->rotate_shutdown)
        break;

      if ($buffer == '')
        sleep(2);
      else
        sleep(250);
    }
//    pclose($eh); //auto cleanup in stop()

    sleep(2); //incase of immediate escape, job is done anyway eh

    $this->idle = true;

    $this->sonic_stream_exec_stop();
  }

  /**
   * log output unifier
   *
   * returns a uniform prefix for log/output entries
   *
   * @see SONIC_stream::sonic_stream_prelog()
   * @return string string usually with pid/process id and timestamp
  */
  function sonic_stream_exec_prelog()
  {
    return $this->sonic_stream_prelog();
  }
}

?>
